@@ -26,14 +26,16 @@ import java.util.Map;
 import java.util.Set;
 
 import org.apache.felix.resolver.ResolverImpl;
+import org.apache.karaf.features.BundleInfo;
 import org.apache.karaf.features.Feature;
 import org.apache.karaf.features.FeaturesService;
+import org.apache.karaf.features.internal.model.Bundle;
 import org.apache.karaf.features.internal.resolver.Slf4jResolverLog;
+import org.apache.karaf.features.internal.service.Deployer;
 import org.apache.karaf.features.internal.service.RepositoryImpl;
 import org.apache.karaf.features.internal.support.TestDownloadManager;
 import org.junit.Test;
 import org.osgi.framework.namespace.IdentityNamespace;
-import org.osgi.framework.wiring.BundleRevision;
 import org.osgi.resource.Capability;
 import org.osgi.resource.Resource;
 import org.osgi.resource.Wire;
@@ -46,28 +48,52 @@ import static org.junit.Assert.assertEquals;
 
 public class SubsystemTest {
 
-    Logger logger = LoggerFactory.getLogger(SubsystemTest.class);;
+    Logger logger = LoggerFactory.getLogger(SubsystemTest.class);
     Resolver resolver = new ResolverImpl(new Slf4jResolverLog(logger));
 
     @Test
     public void test1() throws Exception {
         RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data1/features.xml").toURI());
 
-        Map<String, Set<String>> features = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> features = new HashMap<>();
         addToMapSet(features, "root", "f1");
         addToMapSet(features, "root/apps1", "f2");
 
-        Map<String, Set<String>> expected = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> expected = new HashMap<>();
         addToMapSet(expected, "root", "a/1.0.0");
         addToMapSet(expected, "root", "c/1.0.0");
         addToMapSet(expected, "root/apps1", "b/1.0.0");
 
         SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data1"));
-        resolver.prepare(Arrays.asList(repo.getFeatures()),
+        resolver.prepare(partitionByName(repo.getFeatures()),
                          features,
-                         Collections.<String, Set<BundleRevision>>emptyMap());
-        resolver.resolve(Collections.<String>emptySet(),
-                         FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                         Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                         null, null, null);
+
+        verify(resolver, expected);
+    }
+
+    @Test
+    public void test1a() throws Exception {
+        RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data1/features.xml").toURI());
+
+        Map<String, Set<String>> features = new HashMap<>();
+        addToMapSet(features, "root", "f1");
+        addToMapSet(features, "root/apps1", "f2");
+        addToMapSet(features, "root/apps1/regionx", "bundle:d");
+
+        Map<String, Set<String>> expected = new HashMap<>();
+        addToMapSet(expected, "root", "a/1.0.0");
+        addToMapSet(expected, "root", "c/1.0.0");
+        addToMapSet(expected, "root/apps1", "b/1.0.0");
+        addToMapSet(expected, "root/apps1/regionx", "d/1.0.0");
+
+        SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data1"));
+        resolver.prepare(partitionByName(repo.getFeatures()),
+                         features,
+                         Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
                          null, null, null);
 
         verify(resolver, expected);
@@ -78,12 +104,12 @@ public class SubsystemTest {
 
         RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data2/features.xml").toURI());
 
-        Map<String, Set<String>> features = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> features = new HashMap<>();
         addToMapSet(features, "root/apps1", "f1");
         addToMapSet(features, "root/apps1", "f3");
         addToMapSet(features, "root/apps2", "f1");
 
-        Map<String, Set<String>> expected = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> expected = new HashMap<>();
         addToMapSet(expected, "root/apps1", "c/1.0.0");
         addToMapSet(expected, "root/apps1", "b/1.0.0");
         addToMapSet(expected, "root/apps1", "e/1.0.0");
@@ -94,32 +120,87 @@ public class SubsystemTest {
         addToMapSet(expected, "root/apps2#f1", "a/1.0.0");
 
         SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data2"));
-        resolver.prepare(Arrays.asList(repo.getFeatures()),
+        resolver.prepare(partitionByName(repo.getFeatures()),
                          features,
-                         Collections.<String, Set<BundleRevision>>emptyMap());
-        resolver.resolve(Collections.<String>emptySet(),
-                         FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                         Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
                          null, null, null);
 
         verify(resolver, expected);
     }
 
     @Test
-    public void testOverrides() throws Exception {
+    public void testOverridesCompatibilityModeSymbolicNameMatches() throws Exception {
         RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data3/features.xml").toURI());
 
-        Map<String, Set<String>> features = new HashMap<String, Set<String>>();
+        // this is normally done by features processor
+        ((Bundle)repo.getFeatures()[0].getBundles().get(0)).setOverriden(BundleInfo.BundleOverrideMode.OSGI);
+        ((Bundle)repo.getFeatures()[0].getBundles().get(0)).setOriginalLocation("a");
+        ((Bundle)repo.getFeatures()[0].getBundles().get(0)).setLocation("b");
+
+        Map<String, Set<String>> features = new HashMap<>();
         addToMapSet(features, "root/apps1", "f1");
 
-        Map<String, Set<String>> expected = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> expected = new HashMap<>();
         addToMapSet(expected, "root/apps1", "a/1.0.1");
 
         SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data3"));
-        resolver.prepare(Arrays.asList(repo.getFeatures()),
+        resolver.prepare(partitionByName(repo.getFeatures()),
                          features,
-                         Collections.<String, Set<BundleRevision>>emptyMap());
-        resolver.resolve(Collections.singleton("b"),
-                         FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                         Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                         null, null, null);
+
+        verify(resolver, expected);
+    }
+
+    @Test
+    public void testOverridesCompatibilityModeSymbolicDoesNotMatch() throws Exception {
+        RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data3/features.xml").toURI());
+
+        // this is normally done by features processor
+        ((Bundle)repo.getFeatures()[0].getBundles().get(0)).setOverriden(BundleInfo.BundleOverrideMode.OSGI);
+        ((Bundle)repo.getFeatures()[0].getBundles().get(0)).setOriginalLocation("a");
+        ((Bundle)repo.getFeatures()[0].getBundles().get(0)).setLocation("c");
+
+        Map<String, Set<String>> features = new HashMap<>();
+        addToMapSet(features, "root/apps1", "f1");
+
+        // we expect override not to be used - symbolic name between a and c doesn't match
+        Map<String, Set<String>> expected = new HashMap<>();
+        addToMapSet(expected, "root/apps1", "a/1.0.0");
+
+        SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data3"));
+        resolver.prepare(partitionByName(repo.getFeatures()),
+                         features,
+                         Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                         null, null, null);
+
+        verify(resolver, expected);
+    }
+
+    @Test
+    public void testOverridesMavenMode() throws Exception {
+        RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data3/features.xml").toURI());
+
+        // this is normally done by features processor
+        ((Bundle)repo.getFeatures()[0].getBundles().get(0)).setOverriden(BundleInfo.BundleOverrideMode.MAVEN);
+        ((Bundle)repo.getFeatures()[0].getBundles().get(0)).setOriginalLocation("a");
+        ((Bundle)repo.getFeatures()[0].getBundles().get(0)).setLocation("c");
+
+        Map<String, Set<String>> features = new HashMap<>();
+        addToMapSet(features, "root/apps1", "f1");
+
+        // we expect override to be used - symbolic name between a and c doesn't match, but we don't care
+        Map<String, Set<String>> expected = new HashMap<>();
+        addToMapSet(expected, "root/apps1", "not-a/1.0.1");
+
+        SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data3"));
+        resolver.prepare(partitionByName(repo.getFeatures()),
+                         features,
+                         Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
                          null, null, null);
 
         verify(resolver, expected);
@@ -129,17 +210,16 @@ public class SubsystemTest {
     public void testConditionalUnsatisfiedWithOptional() throws Exception {
         RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data4/features.xml").toURI());
 
-        Map<String, Set<String>> features = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> features = new HashMap<>();
         addToMapSet(features, "root/apps1", "f1");
-        Map<String, Set<String>> expected = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> expected = new HashMap<>();
         addToMapSet(expected, "root/apps1", "a/1.0.0");
 
         SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data4"));
-        resolver.prepare(Arrays.asList(repo.getFeatures()),
+        resolver.prepare(partitionByName(repo.getFeatures()),
                          features,
-                         Collections.<String, Set<BundleRevision>>emptyMap());
-        resolver.resolve(Collections.<String>emptySet(),
-                         FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                         Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
                          null, null, null);
 
         verify(resolver, expected);
@@ -149,19 +229,18 @@ public class SubsystemTest {
     public void testConditionalSatisfiedWithOptional() throws Exception {
         RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data4/features.xml").toURI());
 
-        Map<String, Set<String>> features = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> features = new HashMap<>();
         addToMapSet(features, "root/apps1", "f1");
         addToMapSet(features, "root/apps1", "f2");
-        Map<String, Set<String>> expected = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> expected = new HashMap<>();
         addToMapSet(expected, "root/apps1", "a/1.0.0");
         addToMapSet(expected, "root/apps1", "b/1.0.0");
 
         SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data4"));
-        resolver.prepare(Arrays.asList(repo.getFeatures()),
+        resolver.prepare(partitionByName(repo.getFeatures()),
                          features,
-                         Collections.<String, Set<BundleRevision>>emptyMap());
-        resolver.resolve(Collections.<String>emptySet(),
-                         FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                         Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
                          null, null, null);
 
         verify(resolver, expected);
@@ -171,19 +250,18 @@ public class SubsystemTest {
     public void testBundle() throws Exception {
         RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data1/features.xml").toURI());
 
-        Map<String, Set<String>> features = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> features = new HashMap<>();
         addToMapSet(features, "root/apps1", "bundle:a");
         addToMapSet(features, "root/apps1", "bundle:c;dependency=true");
-        Map<String, Set<String>> expected = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> expected = new HashMap<>();
         addToMapSet(expected, "root/apps1", "a/1.0.0");
         addToMapSet(expected, "root/apps1", "c/1.0.0");
 
         SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data1"));
-        resolver.prepare(Arrays.asList(repo.getFeatures()),
+        resolver.prepare(partitionByName(repo.getFeatures()),
                 features,
-                Collections.<String, Set<BundleRevision>>emptyMap());
-        resolver.resolve(Collections.<String>emptySet(),
-                FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
                 null, null, null);
 
         verify(resolver, expected);
@@ -193,18 +271,17 @@ public class SubsystemTest {
     public void testFeatureOptional() throws Exception {
         RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data5/features.xml").toURI());
 
-        Map<String, Set<String>> features = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> features = new HashMap<>();
         addToMapSet(features, "root", "f1");
-        Map<String, Set<String>> expected = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> expected = new HashMap<>();
         addToMapSet(expected, "root", "a/1.0.0");
         addToMapSet(expected, "root", "b/1.0.0");
 
         SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data5"));
-        resolver.prepare(Arrays.asList(repo.getFeatures()),
+        resolver.prepare(partitionByName(repo.getFeatures()),
                 features,
-                Collections.<String, Set<BundleRevision>>emptyMap());
-        resolver.resolve(Collections.<String>emptySet(),
-                FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
                 null, null, null);
 
         verify(resolver, expected);
@@ -214,19 +291,18 @@ public class SubsystemTest {
     public void testFeatureOptionalAlreadyProvided() throws Exception {
         RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data5/features.xml").toURI());
 
-        Map<String, Set<String>> features = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> features = new HashMap<>();
         addToMapSet(features, "root", "f1");
         addToMapSet(features, "root", "f3");
-        Map<String, Set<String>> expected = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> expected = new HashMap<>();
         addToMapSet(expected, "root", "a/1.0.0");
         addToMapSet(expected, "root", "c/1.0.0");
 
         SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data5"));
-        resolver.prepare(Arrays.asList(repo.getFeatures()),
+        resolver.prepare(partitionByName(repo.getFeatures()),
                 features,
-                Collections.<String, Set<BundleRevision>>emptyMap());
-        resolver.resolve(Collections.<String>emptySet(),
-                FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
                 null, null, null);
 
         verify(resolver, expected);
@@ -236,19 +312,18 @@ public class SubsystemTest {
     public void testFeatureOptionalAlreadyProvided2() throws Exception {
         RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data6/features.xml").toURI());
 
-        Map<String, Set<String>> features = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> features = new HashMap<>();
         addToMapSet(features, "root", "pax-http");
         addToMapSet(features, "root", "pax-http-tomcat");
-        Map<String, Set<String>> expected = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> expected = new HashMap<>();
         addToMapSet(expected, "root", "a/1.0.0");
         addToMapSet(expected, "root", "c/1.0.0");
 
         SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data6"));
-        resolver.prepare(Arrays.asList(repo.getFeatures()),
+        resolver.prepare(partitionByName(repo.getFeatures()),
                 features,
-                Collections.<String, Set<BundleRevision>>emptyMap());
-        resolver.resolve(Collections.<String>emptySet(),
-                FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
                 null, null, null);
 
         verify(resolver, expected);
@@ -258,21 +333,20 @@ public class SubsystemTest {
     public void testResourceRepositories() throws Exception {
         RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data7/features.xml").toURI());
 
-        Map<String, Set<String>> features = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> features = new HashMap<>();
         addToMapSet(features, "root", "f1");
         addToMapSet(features, "root/apps1", "f2");
 
-        Map<String, Set<String>> expected = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> expected = new HashMap<>();
         addToMapSet(expected, "root", "a/1.0.0");
         addToMapSet(expected, "root", "c/1.0.0");
         addToMapSet(expected, "root/apps1", "b/1.0.0");
 
         SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data7"));
-        resolver.prepare(Arrays.asList(repo.getFeatures()),
+        resolver.prepare(partitionByName(repo.getFeatures()),
                 features,
-                Collections.<String, Set<BundleRevision>>emptyMap());
-        resolver.resolve(Collections.<String>emptySet(),
-                FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
                 null, null, null);
 
         verify(resolver, expected);
@@ -295,12 +369,37 @@ public class SubsystemTest {
         addToMapSet(expected, "root", "pax-web-tomcat/6.0.4");
         addToMapSet(expected, "root", "pax-web-api/6.0.4");
 
+        Deployer.DeploymentState ds = new Deployer.DeploymentState();
+        ds.partitionFeatures(allFeatures);
         SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data9"));
-        resolver.prepare(allFeatures,
+        resolver.prepare(ds.featuresByName(),
                 features,
                 Collections.emptyMap());
-        resolver.resolve(Collections.emptySet(),
-                FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
+                null, null, null);
+
+        verify(resolver, expected);
+    }
+
+    @Test
+    public void testBlacklistedBundles() throws Exception {
+        RepositoryImpl repo = new RepositoryImpl(getClass().getResource("data10/features.xml").toURI());
+
+        // this is normally done by features processor
+        ((Bundle)repo.getFeatures()[0].getBundles().get(1)).setBlacklisted(true);
+
+        Map<String, Set<String>> features = new HashMap<>();
+        addToMapSet(features, "root", "f1");
+
+        Map<String, Set<String>> expected = new HashMap<>();
+        // we expect only bundle "a", as "b" is blacklisted
+        addToMapSet(expected, "root", "a/1.0.0");
+
+        SubsystemResolver resolver = new SubsystemResolver(this.resolver, new TestDownloadManager(getClass(), "data10"));
+        resolver.prepare(partitionByName(repo.getFeatures()),
+                features,
+                Collections.emptyMap());
+        resolver.resolve(FeaturesService.DEFAULT_FEATURE_RESOLUTION_RANGE,
                 null, null, null);
 
         verify(resolver, expected);
@@ -327,7 +426,7 @@ public class SubsystemTest {
     }
 
     private Map<String, Set<String>> getBundleNamesPerRegions(SubsystemResolver resolver) {
-        Map<String, Set<String>> mapping = new HashMap<String, Set<String>>();
+        Map<String, Set<String>> mapping = new HashMap<>();
         Map<String, Set<Resource>> bundles = resolver.getBundlesPerRegions();
         for (Map.Entry<String,Set<Resource>> entry : bundles.entrySet()) {
             for (Resource r : entry.getValue()) {
@@ -341,13 +440,8 @@ public class SubsystemTest {
     private void dumpWiring(SubsystemResolver resolver) {
         System.out.println("Wiring");
         Map<Resource, List<Wire>> wiring = resolver.getWiring();
-        List<Resource> resources = new ArrayList<Resource>(wiring.keySet());
-        Collections.sort(resources, new Comparator<Resource>() {
-            @Override
-            public int compare(Resource o1, Resource o2) {
-                return getName(o1).compareTo(getName(o2));
-            }
-        });
+        List<Resource> resources = new ArrayList<>(wiring.keySet());
+        resources.sort(Comparator.comparing(this::getName));
         for (Resource resource : resources) {
             System.out.println("    " + getName(resource));
             for (Wire wire : wiring.get(resource)) {
@@ -363,4 +457,10 @@ public class SubsystemTest {
                 + cap.getAttributes().get(IdentityNamespace.CAPABILITY_VERSION_ATTRIBUTE);
     }
 
+    private Map<String, List<Feature>> partitionByName(Feature[] features) {
+        Deployer.DeploymentState ds = new Deployer.DeploymentState();
+        ds.partitionFeatures(Arrays.asList(features));
+        return ds.featuresByName();
+    }
+
 }
